//=============================================================================
// implementation class midi IO
//=============================================================================
#include <linux/soundcard.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <iostream.h>
#include <mcIO.h>


// define static variables:
const char* TMidiIO::sequencer       = "/dev/sequencer2";
int    TMidiIO::sequencer_fd         = -1;
int    TMidiIO::class_count          =  0;
uchar  TMidiIO::midi_write_packet[4] = {SEQ_MIDIPUTC, 0, 0, 0};
uchar  TMidiIO::midi_read_packet[4];


// static variables for MIDI I/O information database
int    TMidiIO::indevcount      = 0;
int    TMidiIO::outdevcount     = 0;

int*   TMidiIO::indevnum        = NULL;
int*   TMidiIO::outdevnum       = NULL;

int*   TMidiIO::indevtype       = NULL;
int*   TMidiIO::outdevtype      = NULL;

char** TMidiIO::indevnames      = NULL;
char** TMidiIO::outdevnames     = NULL;


//=============================================================================
// constructor
// 04.03.2001 Adrian Meyer
//=============================================================================
TMidiIO::TMidiIO( int autoOpen )
{
  if ( autoOpen ) open();                             // open file descriptor

  if ( GetInputCount() > 0 )
    fInputDevice = 0;
	else
    fInputDevice = -1;

  if ( GetOutputCount() > 0 )
    fOutputDevice = 0;
	else
    fOutputDevice = -1;

  if (class_count < 0) {
     cerr << "Unusual class instatiation count: " << class_count << endl;
     exit(1);
  } else if (class_count == 0) {
     buildInfoDatabase();
  }

  class_count++;
}

//=============================================================================
// destructor
// 04.03.2001 Adrian Meyer
//=============================================================================
TMidiIO::~TMidiIO( void )
{
   class_count--;

   if (class_count == 0) {
      close();
      removeInfoDatabase();
   } else if (class_count < 0) {
      cerr << "Unusual class instatiation count: " << class_count << endl;
      exit(1);
   }
}

//=============================================================================
// read input device
// 05.03.2001 Adrian Meyer
//=============================================================================
int TMidiIO::GetInputDevice( void )
{
  return fInputDevice;
}

//=============================================================================
// set input device
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::SetInputDevice( int aInputDevice )
{
  fInputDevice = aInputDevice;
}

//=============================================================================
// read output device
// 05.03.2001 Adrian Meyer
//=============================================================================
int TMidiIO::GetOutputDevice( void )
{
  return fOutputDevice;
}

//=============================================================================
// set output device
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::SetOutputDevice( int aOutputDevice )
{
  fOutputDevice = aOutputDevice;
}

//=============================================================================
// close the sequencer device
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::close( void )
{
  ::close( getFd() );
}

//=============================================================================
// read input name
// 05.03.2001 Adrian Meyer
//=============================================================================
const char* TMidiIO::GetInputName(int aDevice) const {
   if (aDevice >= GetInputCount()) {
      cerr << "Error: " << aDevice << " is greater than max in ("
           << GetInputCount() << ")" << endl;
      exit(1);
   }

   return indevnames[aDevice];
}

//=============================================================================
// read number of inputs
// 05.03.2001 Adrian Meyer
//=============================================================================
int TMidiIO::GetInputCount(void) const {
   return indevcount;
}

//=============================================================================
// read amount of outputs
// 05.03.2001 Adrian Meyer
//=============================================================================
int TMidiIO::GetOutputCount(void) const {
   return outdevcount;
}

//=============================================================================
// return midi output name
// 05.03.2001 Adrian Meyer
//=============================================================================
const char* TMidiIO::GetOutputName(int aDevice) const {
   if (aDevice >= GetOutputCount()) {
      cerr << "Error: " << aDevice << " is greater than max out ("
           << GetOutputCount() << ")" << endl;
      exit(1);
   }

   return outdevnames[aDevice];
}

//=============================================================================
// is the device open?
// 05.03.2001 Adrian Meyer
//=============================================================================
int TMidiIO::is_open(void) const {
   if (getFd() > 0) {
      return 1;
   } else {
      return 0;
   }
}

//=============================================================================
// open the device
// 05.03.2001 Adrian Meyer
//=============================================================================
int TMidiIO::open(void) {
   if (getFd() <= 0) {
      setFd(::open(sequencer, O_RDWR, 0));
   }

   return is_open();
}

//=============================================================================
// reads MIDI bytes and also stores the
// device from which the byte was read from.  Timing is not
// saved from the device.  If needed, then it would have to
// be saved in this function, or just return the raw packet
// data (use rawread function).
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::read(uchar* buf, uchar* dev, int count) {
   int i = 0;
   while (i < count) {
      ::read(getFd(), midi_read_packet, sizeof(midi_read_packet));
      if (midi_read_packet[1] == SEQ_MIDIPUTC) {
         buf[i] = midi_read_packet[1];
         dev[i] = midi_read_packet[2];
         i++;
      }
   }
}

//=============================================================================
// read Input MIDI packets. each packet contains 4 bytes
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::rawread(uchar* buf, int packetCount) {
   ::read(getFd(), buf, packetCount * 4);
}

//=============================================================================
// rebuild internal database
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::rebuildInfoDatabase(void) {
   removeInfoDatabase();
   buildInfoDatabase();
}

//=============================================================================
// write on ebyte
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::Write(int device, int aByte) {
   switch (getOutputType(device)) {
      case MIDI_EXTERNAL:
         midi_write_packet[1] = (uchar) (0xff & aByte);
         midi_write_packet[2] = getOutDeviceValue(device);
         ::write(getFd(), midi_write_packet, sizeof(midi_write_packet));
         break;
      case MIDI_INTERNAL:
         writeInternal(getOutDeviceValue(device), aByte);
         break;
   }
}

//=============================================================================
// write uchar
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::Write(int device, uchar* bytes, int count) {
   for (int i=0; i<count; i++) {
      Write(device, bytes[i]);
   }
}

//=============================================================================
// write bytes
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::Write(int device, int* bytes, int count) {
   for (int i=0; i<count; i++) {
      Write(device, bytes[i]);
   }
}

//=============================================================================
// build information database of in- and outputs
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::buildInfoDatabase(void) {
   int status;

   // read number of inputs available (external MIDI devices only)
   status = ioctl(getFd(), SNDCTL_SEQ_NRMIDIS, &indevcount);
   if (status!= 0) {
      cerr << "Error determining the number of MIDI inputs" << endl;
      exit(1);
   }

   // read number of output available
   int extmidi = indevcount;
   int intmidi;
   status = ioctl(getFd(), SNDCTL_SEQ_NRSYNTHS, &intmidi);
   if (status!= 0) {
      cerr << "Error determining the number of MIDI inputs" << endl;
      exit(1);
   }
   outdevcount = extmidi + intmidi;

   // allocate space for names and device number arrays
   if (indevnum != NULL || outdevnum != NULL || indevnames != NULL ||
         outdevnames != NULL || indevtype != NULL || outdevtype != NULL) {
      cerr << "Error: buildInfoDatabase called twice." << endl;
      exit(1);
   }

   indevnum = new int[indevcount];
   outdevnum = new int[outdevcount];

   indevtype = new int[indevcount];
   outdevtype = new int[outdevcount];

   indevnames = new char*[indevcount];
   outdevnames = new char*[outdevcount];


   // fill in the device translation table and fill in the device names
   int i;
   struct midi_info midiinfo;
   for (i=0; i<indevcount; i++) {
      midiinfo.device = i;
      status = ioctl(getFd(), SNDCTL_MIDI_INFO, &midiinfo);
      if (status != 0) {
         cerr << "Error while reading MIDI device " << i << endl;
         exit(1);
      }

      indevnum[i]    = midiinfo.device;
      outdevnum[i]   = midiinfo.device;
      indevtype[i]   = MIDI_EXTERNAL;
      outdevtype[i]  = MIDI_EXTERNAL;
      indevnames[i]  = new char[strlen(midiinfo.name) + 1 + 10];
      outdevnames[i] = new char[strlen(midiinfo.name) + 1 + 11];
      strcpy(indevnames[i], midiinfo.name);
      strcpy(outdevnames[i], midiinfo.name);
      strcat(indevnames[i], " (MIDI In)");
      strcat(outdevnames[i], " (MIDI Out)");
   }

   char tempstring[1024] = {0};
   struct synth_info synthinfo;
   for (i=0; i<intmidi; i++) {
      synthinfo.device = i;
      status = ioctl(getFd(), SNDCTL_SYNTH_INFO, &synthinfo);
      if (status != 0) {
         cerr << "Error while reading MIDI device " << i << endl;
         exit(1);
      }
      outdevnum[extmidi+i] = i;
      outdevtype[extmidi + i] = MIDI_INTERNAL;

      strcpy(tempstring, synthinfo.name);
      switch (synthinfo.synth_type) {
         case SYNTH_TYPE_FM:           // 0
            strcat(tempstring, " (FM");
            switch (synthinfo.synth_subtype) {
               case FM_TYPE_ADLIB:     // 0
                  strcat(tempstring, " Adlib");
                  break;
               case FM_TYPE_OPL3:      // 1
                  strcat(tempstring, " OPL3");
                  break;
            }
            strcat(tempstring, ")");
            break;
         case SYNTH_TYPE_SAMPLE:       // 1
            strcat(tempstring, " (Wavetable)");
            break;
         case SYNTH_TYPE_MIDI:         // 2
            strcat(tempstring, " (MIDI Interface");
            switch (synthinfo.synth_subtype) {
               case SYNTH_TYPE_MIDI:   // 0x401
                  strcat(tempstring, " MPU401");
                  break;
            }
            strcat(tempstring, ")");
            break;
      }
      outdevnames[i+extmidi] = new char[strlen(tempstring) + 1];
      strcpy(outdevnames[i+extmidi], tempstring);
   }
}

//=============================================================================
// read file descriptor
// 05.03.2001 Adrian Meyer
//=============================================================================
int TMidiIO::getFd(void) const {
   return sequencer_fd;
}

//=============================================================================
// read input device number
// 05.03.2001 Adrian Meyer
//=============================================================================
int TMidiIO::getInDeviceValue(int aDevice) const {
   if (aDevice >= GetInputCount()) {
      cerr << "Error: " << aDevice << " is greater than max in ("
           << GetInputCount() << ")" << endl;
      exit(1);
   }

   return indevnum[aDevice];
}

//=============================================================================
// returns 1 = external MIDI, 2 = internal MIDI
// 05.03.2001 Adrian Meyer
//=============================================================================
int TMidiIO::getInputType(int aDevice) const {
   if (aDevice >= GetInputCount()) {
      cerr << "Error: " << aDevice << " is greater than max in ("
           << GetInputCount() << ")" << endl;
      exit(1);
   }

   return indevtype[aDevice];
}

//=============================================================================
// read output device number
// 05.03.2001 Adrian Meyer
//=============================================================================
int TMidiIO::getOutDeviceValue(int aDevice) const {
   if (aDevice >= GetOutputCount()) {
      cerr << "Error: " << aDevice << " is greater than max out ("
           << GetOutputCount() << ")" << endl;
      exit(1);
   }

   return outdevnum[aDevice];
}

//=============================================================================
// returns 1 = external MIDI, 2 = internal MIDI
// 05.03.2001 Adrian Meyer
//=============================================================================
int TMidiIO::getOutputType(int aDevice) const {
   if (aDevice >= GetOutputCount()) {
      cerr << "Error: " << aDevice << " is greater than max out ("
           << GetOutputCount() << ")" << endl;
      exit(1);
   }

   return outdevtype[aDevice];
}

//=============================================================================
// kill info database
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::removeInfoDatabase(void) {
   if (indevnum   != NULL)   delete [] indevnum;
   if (outdevnum  != NULL)   delete [] outdevnum;
   if (indevtype  != NULL)   delete [] indevtype;
   if (outdevtype != NULL)   delete [] outdevtype;

   int i;
   if (indevnames != NULL) {
      for (i=0; i<indevcount; i++) {
         if (indevnames[i] != NULL)    delete [] indevnames[i];
      }
      delete [] indevnames;
   }

   if (outdevnames != NULL) {
      for (i=0; i<outdevcount; i++) {
         if (outdevnames[i] != NULL)   delete [] outdevnames[i];
      }
      delete [] outdevnames;
   }

   indevnum    = NULL;
   outdevnum   = NULL;
   indevtype   = NULL;
   outdevtype  = NULL;
   indevnames  = NULL;
   outdevnames = NULL;

   indevcount = 0;
   outdevcount = 0;
}

//=============================================================================
// set the file descriptor
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::setFd( int anFd ) {
   sequencer_fd = anFd;
}

// static variables related to the processing of message for internal MIDI:
uchar  TMidiIO::synth_write_message[8];
uchar  TMidiIO::synth_message_buffer[1024]   = {0};
int    TMidiIO::synth_message_buffer_count   =  0;
int    TMidiIO::synth_message_bytes_expected =  0;
int    TMidiIO::synth_message_curr_device    =  -1;

//=============================================================================
// the device number is the
// driver's device number *NOT* this class's device numbering
// system.  MIDI bytes are stored in a buffer until a complete
// message is received, then a synth message is generated.
// While a complete message is being received, the device number
// cannot change.  The first byte of a message must be a MIDI
// command (i.e., no running status).
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::writeInternal(int aDevice, int aByte) {
   if (synth_message_bytes_expected == 0) {
      // a new message is coming in.
      synth_message_curr_device = aDevice;
      if (aByte < 128) {
         cerr << "Error: MIDI output byte: " << aByte
              << " is not a command byte." << endl;
         exit(1);
      } else {
         synth_message_buffer[0] = aByte;
         synth_message_buffer_count = 1;
      }

      switch (aByte & 0xf0) {
         case 0x80:   synth_message_bytes_expected = 3;   break;
         case 0x90:   synth_message_bytes_expected = 3;   break;
         case 0xA0:   synth_message_bytes_expected = 3;   break;
         case 0xB0:   synth_message_bytes_expected = 3;   break;
         case 0xC0:   synth_message_bytes_expected = 2;   break;
         case 0xD0:   synth_message_bytes_expected = 2;   break;
         case 0xE0:   synth_message_bytes_expected = 3;   break;
         case 0xF0:   cerr << "Can't handle 0xE0 yet" << endl;   exit(1);
         default:     cerr << "Unknown error" << endl;   exit(1);
      }
   }

   // otherwise expecting at least one more byte for the MIDI message
   else {
      if (synth_message_curr_device != aDevice) {
         cerr << "Error: device number changed during message" << endl;
         exit(1);
      }
      if (aByte > 127) {
         cerr << "Error: expecting MIDI data but got MIDI command: "
              << aByte << endl;
         exit(1);
      }

      synth_message_buffer[synth_message_buffer_count++] = aByte;
   }

   // check to see if the message is complete:
   if (synth_message_bytes_expected == synth_message_buffer_count) {
      transmitMessageToInternalSynth();
      synth_message_bytes_expected = 0;
      synth_message_buffer_count = 0;
   }
}

//=============================================================================
// send the stored MIDI message to the internal synthesizer
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::transmitMessageToInternalSynth(void) {
   switch (synth_message_buffer[0] & 0xf0) {
      case 0x80:                      // Note-off
      case 0x90:                      // Note-on
      case 0xA0:                      // Aftertouch
         transmitVoiceMessage();
         break;
      case 0xB0:                      // Control change
      case 0xC0:                      // Patch change
      case 0xD0:                      // Channel pressure
      case 0xE0:                      // Pitch wheel
         transmitCommonMessage();
         break;
      case 0xF0:
         cerr << "Cannot handle 0xf0 commands yet" << endl;
         exit(1);
         break;
      default:
         cerr << "Error: unknown MIDI command" << endl;
         exit(1);
   }
}

//=============================================================================
// send a voice-type MIDI message to an internal synthesizer
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::transmitVoiceMessage(void) {
   synth_write_message[0] = EV_CHN_VOICE;
   synth_write_message[1] = synth_message_curr_device;
   synth_write_message[2] = synth_message_buffer[0] & 0xf0;
   synth_write_message[3] = synth_message_buffer[0] & 0x0f;
   synth_write_message[4] = synth_message_buffer[1];
   synth_write_message[5] = synth_message_buffer[2];
   synth_write_message[6] = 0;
   synth_write_message[7] = 0;

   ::write(getFd(), synth_write_message, sizeof(synth_write_message));
}

//=============================================================================
// send a common-type MIDI message to an internal synthesizer
// 05.03.2001 Adrian Meyer
//=============================================================================
void TMidiIO::transmitCommonMessage(void) {
   synth_write_message[0] = EV_CHN_COMMON;
   synth_write_message[1] = synth_message_curr_device;
   synth_write_message[2] = synth_message_buffer[0] & 0xf0;
   synth_write_message[3] = synth_message_buffer[0] & 0x0f;

   switch (synth_write_message[2]) {
      case 0xB0:                           // Control change
         synth_write_message[4] = synth_message_buffer[1];
         synth_write_message[5] = 0;
         synth_write_message[6] = synth_message_buffer[2];
         synth_write_message[7] = 0;
         break;
      case 0xC0:                           // Patch change
      case 0xD0:                           // Channel pressure
         synth_write_message[4] = synth_message_buffer[1];
         synth_write_message[5] = 0;
         synth_write_message[6] = 0;
         synth_write_message[7] = 0;
         break;
      case 0xE0:                           // Pitch wheel
         synth_write_message[4] = 0;
         synth_write_message[5] = 0;
         synth_write_message[6] = synth_message_buffer[1];
         synth_write_message[7] = synth_message_buffer[2];
         break;
      default:
         cerr << "Unknown Common MIDI message" << endl;
         exit(1);
   }

   ::write(getFd(), synth_write_message, sizeof(synth_write_message));
}